import cv2
import cv2.cv as cv
import numpy as np
import matplotlib.pyplot as plt
import scipy.io
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mlabwrap import mlab
import pickle

#import time

def detection(image):

	# import image
	frame_cv = image
	frame_np = np.array(frame_cv[:,:])

	# convert to YUV
	yuv_np = cv2.cvtColor(frame_np, cv.CV_BGR2YCrCb)
	yuv_cv = cv.fromarray(yuv_np)

	colors = ['green','yellow','blue','red']
	lowerb = [cv.Scalar(1,160,100), cv.Scalar(1,100,1), cv.Scalar(1,50,100), cv.Scalar(1,100,140)]
	upperb = [cv.Scalar(255,255,150), cv.Scalar(255,150,107), cv.Scalar(255,110,150), cv.Scalar(100,150,200)]
	centers = []

	for i in range(0,4):

		# threshold for color of interest
		color=cv.CreateImage(cv.GetSize(frame_cv), cv.IPL_DEPTH_8U, 1)
		cv.InRangeS(yuv_cv,lowerb[i],upperb[i], color)
		color_np = np.array(color[:,:]) #necessary later
	
		# get rid of noise
		cv.Erode(color,color,None,1)
		cv.Dilate(color,color,None,1)
		cv.Dilate(color,color,None,3)
		cv.Erode(color,color,None,3)

		#choose maximum contour
		contours, hierarchy = cv2.findContours(color_np, cv2.RETR_LIST,cv2.CHAIN_APPROX_SIMPLE)
		max_area = 0
		for cnt in contours:
			area = cv2.contourArea(cnt)
			if area > max_area:
				max_area = area
				best_cnt = cnt
		#print(max_area)
	
		if max_area > 0:
			#find centers
			M = cv2.moments(best_cnt)
			cx,cy = int(M['m10']/M['m00']), int(M['m01']/M['m00'])
			centers.append([cx, cy])
	
		else:
			centers.append(-1)

	return centers





def tracking(kc1, kc2, kc3, kc4, KK1, KK2, KK3, KK4, R2to1, R3to2, R4to3, T2to1, T3to2, T4to3, centers1,centers2,centers3,centers4):

        points = np.array([])
        X0 = np.array([0.,0.,0.])

        #for each marker
        for i in range(4):

                #check if marker exists
                check = [];
                XX1 = np.array([[-1.,-1.,-1.],[-1.,-1.,-1.],[-1.,-1.,-1.],[-1.,-1.,-1.]])
                XX0 = np.array([[-1.,-1.,-1.],[-1.,-1.,-1.],[-1.,-1.,-1.],[-1.,-1.,-1.]])
            
                if type(centers1[i]) != int:
                        check.append(0)                    
                        #find image to camera coordinates
                        #option one: ignore distortion
                        #x = np.append(centers1[i],1)
                        #X1 = np.dot(np.linalg.inv(KK1),x) #Y=KK*X solve for x
                    
                        #option two: account for distortion
                        x = np.asarray(centers1[i],dtype=np.float)
                        X1 = mlab.normalize(x, np.array([KK1[0][0], KK1[1][1]]), np.array([KK1[0][2], KK1[1][2]]), kc1, np.array(KK1[0][1]/KK1[0,0])) #x_pixel, fc, cc, kc, alpha_c
                        X1 = np.append(X1,1.)

                        #transform to camera 1
                        XX1[0] = X1
                        XX0[0] = X0

                if type(centers2[i]) != int:
                        check.append(1)
                        
                        #x = np.append(centers2[i],1)
                        #X2 = np.dot(np.linalg.inv(KK2),x)

                        x = np.asarray(centers2[i],dtype=np.float)
                        X2 = mlab.normalize(x, np.array([KK2[0][0], KK2[1][1]]), np.array([KK2[0][2], KK2[1][2]]), kc2, np.array(KK2[0][1]/KK2[0,0])) #x_pixel, fc, cc, kc, alpha_c
                        X2 = np.append(X2,1.)
                        
                        XX1[1] = np.dot(R2to1,X2)+T2to1.T[0]
                        XX0[1] = np.dot(R2to1,X0)+T2to1.T[0]
                    
                if len(centers3[i]) > 1:
                        check.append(2)
                        
                        #x = np.append(centers3[i],1)
                        #X3 = np.dot(np.linalg.inv(KK3),x)

                        x = np.asarray(centers3[i],dtype=np.float)
                        X3 = mlab.normalize(x, np.array([KK3[0][0], KK3[1][1]]), np.array([KK3[0][2], KK3[1][2]]), kc2, np.array(KK3[0][1]/KK3[0,0])) #x_pixel, fc, cc, kc, alpha_c
                        X3 = np.append(X3,1.)

                        XX1[2] = np.dot(R2to1,(np.dot(R3to2,X3)+T3to2.T[0]))+T2to1.T[0]
                        XX0[2] = np.dot(R2to1,(np.dot(R3to2,X0)+T3to2.T[0]))+T2to1.T[0]
            
                if type(centers4[i]) != 1:
                        check.append(3)
                        #x = np.append(centers4[i],1)
                        #X4 = np.dot(np.linalg.inv(KK4),x)

                        x = np.asarray(centers4[i],dtype=np.float)
                        X4 = mlab.normalize(x, np.array([KK4[0][0], KK4[1][1]]), np.array([KK4[0][2], KK4[1][2]]), kc4, np.array(KK4[0][1]/KK4[0,0])) #x_pixel, fc, cc, kc, alpha_c
                        X4 = np.append(X4,1.)

                        XX1[3] = np.dot(R2to1,(np.dot(R3to2,(np.dot(R4to3,X4)+T4to3.T[0])+T3to2.T[0])))+T2to1.T[0]
                        XX0[3] = np.dot(R2to1,(np.dot(R3to2,(np.dot(R4to3,X0)+T4to3.T[0])+T3to2.T[0])))+T2to1.T[0]


                #find point
                n = len(check)
                if n == 2: #two camera views of marker
                        i = check[0]
                        j = check[1]
                        pt = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        points = np.append(points, pt)
                elif n == 3: #3 camera views of marker
                        i = check[0]
                        j = check[1]
                        pt1 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        i = check[1]
                        j = check[2]
                        pt2 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        i = check[0]
                        j = check[2]
                        pt3 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        pt = np.divide((pt1 + pt2 + pt3),3.0)
                        points = np.append(points, pt)
                elif n == 4: #4 camera views of marker
                        i = check[0]
                        j = check[1]
                        pt1 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])
                    
                        i = check[0]
                        j = check[2]
                        pt2 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])
                    
                        i = check[0]
                        j = check[3]
                        pt3 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        i = check[1]
                        j = check[2]
                        pt4 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        i = check[1]
                        j = check[3]
                        pt5 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        i = check[2]
                        j = check[3]
                        pt6 = closestpt(XX0[i],XX1[i],XX0[j],XX1[j])

                        pt = np.divide((pt1 + pt2 + pt3 + pt4 + pt5 +pt6),6.0)
                        points = np.append(points, pt)
                else: #not enough camera views
                        print "No marker in 3 or more camera views."
                        #points = np.append(points, -1)

	#return pt with respect to camera 1
        points = np.reshape(points, (len(points)/3,3))
        return points


def closestpt(x00, x01, x10, x11):
        #common shortest distance algorithm
        slope0 = x01-x00
        slope1 = x11-x10
        x = np.cross(slope0,slope1)
        u = np.divide(x,np.sqrt(np.dot(x,x)))
        g = np.dot(x00-x10,u)
        
        #least squares, SVD
        b = -g*u-x00+x10
        A = np.array([[slope0[0],-slope1[0]],[slope0[1],-slope1[1]],[slope0[2],-slope1[2]]])
        U,s,Vt = np.linalg.svd(A)
        Srecip = np.divide(1,s)
        Sinv = np.array([[Srecip[0],0,0],[0,Srecip[1],0]])
        X = np.dot(np.dot(np.dot(np.transpose(Vt),Sinv),np.transpose(U)),b)

        #average of points with smallest distance between lines
        pt1 = x00 + slope0*X[0]
        pt2 = x10 + slope1*X[1]
        pt = np.divide(pt1+pt2,2.0)

        return pt


def updateDisplay(points):
        n=30

        #reset figure
        ax.cla() #clear figure
        ax.set_xlabel('X (mm)')
        ax.set_ylabel('Y (mm)')
        ax.set_zlabel('Z (mm)')

        if len(points) == 4:

                color = ['r','y','g','b']

                #draw points
                for i in range(4):
                        ax.scatter(points[i][0],points[i][1],points[i][2], s=50, c=color[i], marker = 'o') 

                #draw line
                L = 20 ########################## in mm
                x0 = points[3][0]
                y0 = points[3][1]
                z0 = points[3][2]
                slopex = (points[1][0]-points[3][0])
                slopey = (points[1][1]-points[3][1])
                slopez = (points[1][2]-points[3][2])
                u = np.array([slopex, slopey, slopez])
                slopexNORM = slopex/np.sqrt(np.dot(u,u))
                slopeyNORM = slopey/np.sqrt(np.dot(u,u))
                slopezNORM = slopez/np.sqrt(np.dot(u,u))
                x = slopexNORM*L + x0
                y = slopeyNORM*L + y0
                z = slopezNORM*L + z0
                ax.plot([x0,x],[y0,y],[z0,z],label='tool shaft') #plot a line
                ax.legend()

                #draw plane
                point = points[0]
                normal = np.cross(points[1]-points[0],points[2]-points[0]) 
                d = -np.sum(point*normal)
                xx, yy = np.meshgrid(range(10,30),range(-20,5)) ######### change as needed
                z = (-normal[0]*xx - normal[1]*yy - d)/normal[2]
                ax.plot_surface(xx,yy,z, alpha=0.1)
                
        else:
                print 'There are not enough markers detected to render a drawing.'


        #set bounds and draw
        ax.set_xlim(-n,n)
        ax.set_ylim(-n,n)
        ax.set_zlim(80,120) ########### fix this as necessary
        fig.canvas.draw() #update the plot

        return


#main

#initialize matlab
mlab.addpath('/Users/Sue/Documents/MATLAB/TOOLBOX_calib/')

#initialize display
fig = plt.figure()
ax = fig.gca(projection='3d')

#read in calibration (matlab) variables #######################################################
intrins1 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera1/Calib_Results.mat')
intrins2 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera2/Calib_Results.mat')
intrins3 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera3/Calib_Results.mat')
intrins4 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera4/Calib_Results.mat')
extrins1 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera1v2/Calib_Results_stereo.mat')
extrins2 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera2v3/Calib_Results_stereo.mat')
extrins3 = scipy.io.loadmat('/Users/Sue/Dropbox/CIS2/Code/Calibration/Camera3v4/Calib_Results_stereo.mat')

#extract relevant parameters
KK1 = intrins1['KK']
KK2 = intrins2['KK']
KK3 = intrins3['KK']
KK4 = intrins4['KK']
R2to1 = extrins1['R']
T2to1 = extrins1['T']
R3to2 = extrins2['R']
T3to2 = extrins2['T']
R4to3 = extrins3['R']
T4to3 = extrins3['T']

#option 2
kc1 = intrins1['kc']
kc2 = intrins2['kc']
kc3 = intrins3['kc']
kc4 = intrins4['kc']


#detection
path = "/Users/Sue/Desktop/CIS2/TrackingEval/Static/trial1/light1/" ########################################

image = cv.LoadImageM(path+"cam1_1.bmp") ###############################################
centers1 = detection(image)

image = cv.LoadImageM(path+"cam2_1.bmp") ###############################################
centers2 = detection(image)

image = cv.LoadImageM(path+"cam3_1.bmp") ###############################################
centers3 = detection(image)

image = cv.LoadImageM(path+"cam4_1.bmp") ###############################################
centers4 = detection(image)


#tracking
points = tracking(kc1, kc2, kc3, kc4, KK1, KK2, KK3, KK4, R2to1, R3to2, R4to3, T2to1, T3to2, T4to3, centers1, centers2, centers3, centers4) #######################################

#update display
updateDisplay(points)

#save and clean up
print points
f = open(path+'points1','w') #####################################################################
pickle.dump(points, f)
f.close()

fig.savefig(path+"result1.png") ####################################################################
cv2.destroyAllWindows()
